# STL
import re
from abc import ABC, abstractmethod

# PDM
from typing_extensions import override


class Cleaner(ABC):
    @classmethod
    @abstractmethod
    def clean(cls, token: str) -> str:
        """Transform a token to remove some undesirable part."""
        raise NotImplementedError


class RegexCleaner(Cleaner):
    pattern: "re.Pattern[str]"
    replace: str

    @classmethod
    @override
    def clean(cls, token: str) -> str:
        return re.sub(cls.pattern, cls.replace, token)


class ConsecutiveDuplicates(Cleaner):
    """Remove consecutive duplicates from an input string, ignoring case.

    The first match of any 2+ will become `\\1`, preserving initial case.
    For example, `FfFoo` will reduce to `Foo`, and `bBAR` will reduce to `bAR`.

    This is desirable for Toki Pona written with the Latin alphabet because strings
    may be altered for emphasis or effect, such as in "sonaaaa" or "AAAAAA".

    This may be undesirable for moraic scripts like Hiragana, where `わわ` would be
    incorrectly reduced to `わ`. This does preserve phonotactic validity, though.
    """

    @classmethod
    @override
    def clean(cls, token: str) -> str:
        if not token:
            return token

        output = token[0]

        last_output = output.lower()  # ignore case in comparison
        for i in range(1, len(token)):
            cur_char = token[i].lower()
            if cur_char == last_output:
                continue
            output += token[i]  # preserve case of string
            last_output = cur_char
        return output


class ConsecutiveDuplicatesRe(RegexCleaner):
    """Reference implementation for `ConsecutiveDuplicates`."""

    pattern = re.compile(r"(.)\1+", flags=re.IGNORECASE)
    replace = r"\1"


class Lowercase(Cleaner):
    @classmethod
    @override
    def clean(cls, token: str) -> str:
        return token.lower()


__all__ = [
    "ConsecutiveDuplicates",
    "Lowercase",
]



# STL
from typing import Optional

# PDM
import pytest
import hypothesis.strategies as st
from hypothesis import given, example

# LOCAL
from sonatoki.Preprocessors import (
    URLs,
    Spoilers,
    AllQuotes,
    Backticks,
    Codeblock,
    Reference,
    ArrowQuote,
    ColonEmotes,
    DoubleQuotes,
    MarkdownURLs,
    SingleQuotes,
    DiscordEmotes,
    DiscordSpecial,
    DiscordChannels,
    DiscordMentions,
    AngleBracketObject,
)
from src.sonatoki.Preprocessors import Emails


def extract_bracket_content(markdown_text: str) -> Optional[str]:
    start = markdown_text.find("[", 0)
    if start == -1:
        return None

    end = markdown_text.rfind("](")
    if end == -1 or end <= start:
        return None

    return markdown_text[start + 1 : end]


@given(st.from_regex(URLs.pattern, fullmatch=True))
@example("https://google.com")
@example("https://mun.la")
@example("https://discord.gg/")
@example("http://example.com")
@example("http://localhost:80")
def test_URLs(s: str):
    assert URLs.process(s).strip() == ""


@given(st.from_regex(MarkdownURLs.pattern, fullmatch=True))
@example("[mu](https://google.com)")
@example("[wawa wawa wawa](https://mun.la)")
@example("[[] silly mode activated](https://discord.gg/)")
@example("[https://example.com/](http://example.com)")
@example("[192.168.0.255](http://localhost:80)")
@example("[text](https://bad.worse]/)")
@example("[](](http://0)")
def test_MarkdownURLs(s: str):
    bracket_content = extract_bracket_content(s)
    assert MarkdownURLs.process(s) == bracket_content


@given(st.from_regex(Emails.pattern, fullmatch=True))
@example("mun@pona.la")
@example("tokipona@alinome.com")
def test_Emails(s: str):
    assert Emails.process(s).strip() == ""


@given(st.from_regex(Spoilers.pattern, fullmatch=True))
@example("|| | ||")
@example("|| content\n\n\ncontent ||")
@example("||\n||")
@example("|| \n|\n\n || || ||")
def test_Spoilers(s: str):
    res = Spoilers.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(Backticks.pattern, fullmatch=True))
@example("` ` ` `")
def test_Backticks(s: str):
    res = Backticks.process(s).strip()
    assert res == "", (repr(s), repr(res))


@pytest.mark.skip("it observably works but my test for that is inaccurate")
@given(st.from_regex(r"```(?:(?!`).+?)```", fullmatch=True))
@example("""```0```""")
@example(
    """```
```"""
)
@example(
    """```
blocky message
```

```
second blocky message
```"""
)
@example(
    """```oisandm123-_mu
arbitrary content
```"""
)
@example("""```mu```""")
def test_Codeblock(s: str):
    res = Codeblock.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(ArrowQuote.pattern, fullmatch=True))
@example("> base")
@example("> newline\n> newline")
def test_ArrowQuote(s: str):
    res = ArrowQuote.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DoubleQuotes.pattern, fullmatch=True))
@example('" "" "')
@example('" "\n" "')
@example('" \n "')
def test_DoubleQuotes(s: str):
    res = DoubleQuotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(SingleQuotes.pattern, fullmatch=True))
@example("' '' '")
@example("' '\n' '")
@example("' \n '")
def test_SingleQuotes(s: str):
    res = SingleQuotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordEmotes.pattern, fullmatch=True))
@example("<a:example:123123>")
@example("<:example:123123>")
def test_DiscordEmotes(s: str):
    res = DiscordEmotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordMentions.pattern, fullmatch=True))
@example("<@497549183847497739>")
@example("<@!457890000>")
@example("<@&18398198981985>")
def test_DiscordMentions(s: str):
    res = DiscordMentions.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordChannels.pattern, fullmatch=True))
@example("<#19858915>")
@example("<#18591912589812985>")
def test_DiscordChannels(s: str):
    res = DiscordChannels.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(DiscordSpecial.pattern, fullmatch=True))
@example("<id:guide>")
@example("<id:browse>")
def test_DiscordSpecial(s: str):
    res = DiscordSpecial.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(
    st.from_regex(DiscordEmotes.pattern, fullmatch=True)
    | st.from_regex(DiscordMentions.pattern, fullmatch=True)
    | st.from_regex(DiscordChannels.pattern, fullmatch=True)
    | st.from_regex(DiscordSpecial.pattern, fullmatch=True)
    | st.from_regex(AngleBracketObject.pattern, fullmatch=True)
)
@example("<https://example.com>")
@example("<#123124125125>")
def test_AngleBracketObject(s: str):
    res = AngleBracketObject.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(
    st.from_regex(SingleQuotes.pattern, fullmatch=True)
    | st.from_regex(DoubleQuotes.pattern, fullmatch=True)
    | st.from_regex(Backticks.pattern, fullmatch=True)
    | st.from_regex(ArrowQuote.pattern, fullmatch=True)
    | st.from_regex(AllQuotes.pattern, fullmatch=True)
)
@example("> bruh")
@example("`bruh`")
def test_AllQuotes(s: str):
    res = AllQuotes.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(Reference.pattern, fullmatch=True))
@example("[[Brainstorm]]")
@example("[[Phatic Phrases]]")
@example("[[Yahoo!]]")
def test_Reference(s: str):
    res = Reference.process(s).strip()
    assert res == "", (repr(s), repr(res))


@given(st.from_regex(ColonEmotes.pattern, fullmatch=True))
@example(":owe::owe:")
@example(":suffering:")
@example(":presid65despair:")
def test_ColonEmotes(s: str):
    res = ColonEmotes.process(s).strip()
    assert res == "", (repr(s), repr(res))



# PDM
import hypothesis.strategies as st

# LOCAL
from sonatoki.Filters import Syllabic, Phonotactic, AlphabeticRe
from sonatoki.constants import words_by_usage

PROPER_NAME_RE = r"[A-Z][a-z]*"

token_strategy = (
    st.sampled_from(list(words_by_usage(60)))
    | st.from_regex(Phonotactic.pattern, fullmatch=True)
    | st.from_regex(Syllabic.pattern, fullmatch=True)
    | st.from_regex(PROPER_NAME_RE, fullmatch=True)
    | st.from_regex(AlphabeticRe.pattern, fullmatch=True)
)


token_list_strategy = st.lists(
    token_strategy,
    min_size=0,
    max_size=10,
    unique=True,
)



# STL
from typing import List, Type

# PDM
import pytest
import hypothesis.strategies as st
from hypothesis import given, example

# LOCAL
from sonatoki.Filters import (
    Filter,
    NimiPu,
    PuName,
    Numeric,
    Syllabic,
    Alphabetic,
    Phonotactic,
    NimiLinkuCore,
    PunctuationRe,
    NimiLinkuCommon,
)
from sonatoki.Scorers import (
    Scorer,
    Voting,
    Scaling,
    PassFail,
    SoftVoting,
    SoftScaling,
    SoftPassFail,
)

# FILESYSTEM
from .test_utils import token_strategy

FILTERS = [
    NimiPu,
    Numeric,
    Syllabic,
    NimiLinkuCore,
    NimiLinkuCommon,
    Alphabetic,
    PuName,
    Phonotactic,
    PunctuationRe,
]

SCORERS = [
    PassFail,
    SoftPassFail,
    Scaling,
    SoftScaling,
    Voting,
    SoftVoting,
]


@pytest.mark.parametrize("scorer", SCORERS)  # test each scorer
@given(
    st.lists(st.sampled_from(FILTERS), min_size=1, unique=True),
    st.lists(token_strategy, min_size=0, max_size=10),
)
@example(st.sampled_from(FILTERS), [])
def test_score_bounds(scorer: Scorer, filters: List[Type[Filter]], text: List[str]):
    score = scorer.score(text, filters)
    assert 0 <= score <= 1, (score, filters, text)



# STL
from typing import List, Type, TypedDict

# PDM
from typing_extensions import NotRequired

# LOCAL
from sonatoki.types import Number
from sonatoki.Filters import (
    Or,
    And,
    Len,
    Not,
    Filter,
    PuName,
    Numeric,
    Syllabic,
    NimiUCSUR,
    Alphabetic,
    NimiKuLili,
    NimiKuSuli,
    ProperName,
    Punctuation,
    LongSyllabic,
    Miscellaneous,
    LongAlphabetic,
    LongProperName,
    FalsePosSyllabic,
    NimiLinkuByUsage,
    NimiLinkuObscure,
    NimiLinkuSandbox,
    NimiLinkuUncommon,
    FalsePosAlphabetic,
)
from sonatoki.Scorers import Scorer, Soften, Voting, PassFail, SoftScaling, SoftPassFail
from sonatoki.Cleaners import Cleaner, ConsecutiveDuplicates
from sonatoki.Tokenizers import Tokenizer, WordTokenizerRe
from sonatoki.Preprocessors import (
    RECOMMENDED_PREPROCESSORS,
    URLs,
    Emoji,
    Codeblock,
    Reference,
    Preprocessor,
    AngleBracketObject,
)

__DICT_PHONOMATCHES = {
    # Sandbox words are removed from the CorpusConfig if they appear more frequently in English than Toki Pona by a factor of at least 3.
    # In this case, all of these appear more often in English by a factor of at least 10.
    "aka",  # also known as
    "an",  # article
    "api",  # API
    "i",  # 1st person
    "je",  # 1st person pronoun, french
    "kana",  # japanese script
    "me",  # 1st person singular, english
    "ne",  # "no" in several languages
    "nu",  # "new" in english, "now" in dutch
    "omen",  # ominous
    "se",  # spanish particle, english "see"
    "sole",  # singular, of shoe
    "take",  # acquire, perhaps forcefully or without permission
    "ten",  # 10
    "to",  # to, too
    "u",  # no u
    "we",  # 1st person plural, english
    "wi",  # wii and discussions of syllables
    # unexplored candidates for removal
    # "papa",  # father
    # "lo",  # "lo" and "loo"
    # "ewe",  # sheep
    # "pa",  # father- eh?
}


class IloConfig(TypedDict):
    preprocessors: List[Type[Preprocessor]]
    cleaners: List[Type[Cleaner]]
    ignoring_filters: List[Type[Filter]]
    scoring_filters: List[Type[Filter]]
    scorer: Type[Scorer]
    passing_score: Number
    word_tokenizer: NotRequired[Type[Tokenizer]]
    sent_tokenizer: NotRequired[Type[Tokenizer]]


# TODO: branching configs? config builder?

BaseConfig: IloConfig = {
    "preprocessors": [URLs],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [],
    "scorer": PassFail,
    "passing_score": 0.8,
}


PrefConfig: IloConfig = {
    "preprocessors": RECOMMENDED_PREPROCESSORS,
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        Len(Or(NimiLinkuByUsage(30), NimiUCSUR), max=15),
        Len(And(Syllabic, Not(FalsePosSyllabic)), min=3, max=24),
        # NOTE: These are allowed to pass name and alphabetic below, because they *could* be wrong
        Len(ProperName, min=2, max=24),
        Len(And(Alphabetic, Not(FalsePosAlphabetic)), min=3, max=24),
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}

CorpusConfig: IloConfig = {
    "preprocessors": RECOMMENDED_PREPROCESSORS,
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        Len(
            Or(
                # awkward but efficient syntax
                NimiLinkuByUsage(0)(sub=__DICT_PHONOMATCHES),
                NimiUCSUR,
                Miscellaneous,
            ),
            max=19,
        ),
        Len(And(Syllabic, Not(FalsePosSyllabic)), min=3, max=24),
        Len(ProperName, min=2, max=24),
        Len(And(Alphabetic, Not(FalsePosAlphabetic)), min=3, max=24),
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}
"""Mimics the previous implementation of ilo pi toki pona taso."""
LazyConfig: IloConfig = {
    "preprocessors": RECOMMENDED_PREPROCESSORS,
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [Alphabetic, NimiUCSUR, PuName, Miscellaneous],
    "scorer": SoftPassFail,
    "passing_score": 0.8,
    "word_tokenizer": WordTokenizerRe,  # mimics old tokenizer
}
"""This is extremely silly."""
IsipinEpikuConfig: IloConfig = {
    "preprocessors": RECOMMENDED_PREPROCESSORS,
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [
        Or(
            NimiKuSuli,
            NimiKuLili,
            NimiLinkuUncommon,
            NimiLinkuObscure,
            NimiLinkuSandbox,
        ),
        And(LongSyllabic, Not(FalsePosSyllabic)),
        LongProperName,
        And(LongAlphabetic, Not(FalsePosAlphabetic)),
    ],
    "scorer": SoftScaling,
    "passing_score": 0.8,
}


__all__ = [
    "BaseConfig",
    "CorpusConfig",
    "IloConfig",
    "LazyConfig",
    "PrefConfig",
]



# STL
import math
from abc import ABC, abstractmethod
from typing import List, Type

# PDM
from typing_extensions import override

# LOCAL
from sonatoki.types import Number, Scorecard
from sonatoki.Filters import Pass, Filter


class Scorer(ABC):
    @classmethod
    @abstractmethod
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        """Score a list of tokens using the given `Filter`s, returning a
        `Number` between 0 and 1 inclusive."""
        raise NotImplementedError


class Soften(Scorer):
    """Meta `Scorer` which scales the scores of short messages to reduce the
    impact of shortness on scoring.

    The scores of short messages are scaled by mapping the token count
    to [0.5, 1.0] via the sigmoid function, then raising the score to
    the resultant power.

    For example, a single token scoring 0.64 will score 0.8 instead.
    """

    @staticmethod
    def sigmoid(n: int) -> Number:
        return 1 / (1 + math.exp(-(0.30 * (n - 1))))
        # n-1 makes sigmoid(1) == 0.5
        # 0.30 softens scaling in favor of short input
        # return n / (1+abs(n))   # too weak in 0.7+

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        percentage = super().score(tokens, filters)  # type: ignore [abstractmethod]
        len_tokens = len(tokens)
        percentage **= cls.sigmoid(len_tokens)
        return percentage

    def __new__(cls, scorer: Type[Scorer]) -> Type[Scorer]:
        class SoftenedScorer(Soften, scorer): ...

        return SoftenedScorer


class PassFail(Scorer):
    """If a token matches any filter, it scores 1.

    Otherwise, it scores 0.
    """

    @classmethod
    def score_token(cls, token: str, filters: List[Type[Filter]]) -> Number:
        for f in filters:
            if f.filter(token):
                return 1
        return 0

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_tokens = len(tokens)
        for token in tokens:
            total_score += cls.score_token(token, filters)
        return total_score / len_tokens if len_tokens else 0


class Scaling(Scorer):
    """Tokens score 1 for matching the first filter, and a linearly reduced
    amount for matching later filters based on how many filters there are.

    For example, if there are 4 filters, a token scores 1.0, 0.75, 0.50,
    and 0.25 for matching each respectively.

    In other words, filter order matters, weighing earlier listed
    filters higher than later ones. This is desirable to avoid messages
    which would only match weaker filters, as these are less likely to
    be Toki Pona.
    """

    @classmethod
    def score_token(cls, token: str, filters: List[Type[Filter]], scale: int):
        for i, f in enumerate(filters):
            if f.filter(token):
                return scale - i
        return 0

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_filters = len(filters)
        max_score = len(tokens) * len_filters
        for token in tokens:
            total_score += cls.score_token(token, filters, len_filters)
        return total_score / max_score if max_score else 0


class Voting(Scaling):
    """Derives from `Scaling` in assigning scores from 0 to 1 based on the
    first matching filter out of the list of filters. However, after all scores
    are derived, each token scoring less than the threshold is assigned the
    average score of its nearest 3 neighbors. The default threshold is 0.

    If there are 3 or fewer tokens, this scorer is identical to the
    Scaling scorer.

    If the Voting scorer is created with a Filter, tokens must also
    match that filter to be considered for voting. For example, the
    following Voting filter would only check words with a score of 0.3
    or less that still match the Syllabic filter: `Voting(Syllabic, 0.3)`
    """

    prereq: Type[Filter] = Pass
    threshold: int = 0

    def __new__(cls, filter: Type[Filter], threshold_: int = 0) -> Type[Scorer]:
        class AnonVoting(Voting):
            prereq = filter
            threshold = threshold_

        return AnonVoting

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        if len(tokens) < 4:
            return super().score(tokens, filters)

        len_filters = len(filters)
        max_score = len(tokens) * len_filters

        # score_token only emits ints
        # but the averaging emits floats
        # it doesn't really matter as long as no score exceeds len_filters
        scores: List[Number] = []
        for token in tokens:
            score = cls.score_token(token, filters, len_filters)
            scores.append(score)

        # only consider scores from before voting
        copied_scores = scores[:]
        for i, (token, score) in enumerate(zip(tokens, copied_scores)):
            if score > cls.threshold:
                continue
            if not cls.prereq.filter(token):
                continue

            # TODO: this is kinda dumb.
            # we want to get exactly 3 neighbors, favoring 2 before and 1 after
            # the way i'm doing this is both bad and slow as hell
            start = max(i - 2, 0)
            end = min(i + 1, len(scores) - 1)
            neighbors = copied_scores[start:i] + copied_scores[i + 1 : end + 1]
            scores[i] = sum(neighbors) / len(neighbors)

        total_score = sum(scores)

        return total_score / max_score if max_score else 0


class SoftPassFail(Soften, PassFail):
    """Same as `PassFail`, but shorter messages are subject to less harsh
    scoring."""


class SoftScaling(Soften, Scaling):
    """Same as `Scaling`, but shorter messages are subject to less harsh
    scoring."""


class SoftVoting(Soften, Voting):
    """Same as `Voting`, but shorter messages are subject to less harsh
    scoring."""


class SentenceScorer(ABC):
    @classmethod
    @abstractmethod
    def score(cls, scorecards: List[Scorecard]) -> List[Scorecard]:
        """Re-score a list of sentences (scorecards, sentences with all their
        metadata) and return them."""
        raise NotImplementedError


class SentNoOp(SentenceScorer):
    @classmethod
    @override
    def score(cls, scorecards: List[Scorecard]) -> List[Scorecard]:
        return scorecards


class SentAvg(SentenceScorer):
    @classmethod
    @override
    def score(cls, scorecards: List[Scorecard]) -> List[Scorecard]:
        if not scorecards:
            return scorecards

        total = sum(card["score"] for card in scorecards)
        avg = total / len(scorecards)
        for card in scorecards:
            card["score"] = avg
        return scorecards


class SentWeightedAvg(SentenceScorer):
    @classmethod
    @override
    def score(cls, scorecards: List[Scorecard]) -> List[Scorecard]:
        if not scorecards:
            return scorecards

        weighted_total = 0
        total_len = 0
        for card in scorecards:
            cardlen = len(card["cleaned"])
            cardscore = card["score"]

            weighted_total += cardlen * cardscore
            total_len += cardlen

        weighted_avg = weighted_total / total_len
        for card in scorecards:
            card["score"] = weighted_avg
        return scorecards


__all__ = [
    "PassFail",
    "Scaling",
    "SoftPassFail",
    "SoftScaling",
    "Soften",
    "SentAvg",
    "SentWeightedAvg",
]



"""
"Preprocessors" are classes which strip content from a given string prior to tokenization.
There are currently two distinct types of Preprocessor:

- Remove a token from a string which would be difficult to identify after tokenization.
  - URLs
  - DiscordEmotes
- Remove a section of a string which is contained in or marked by certain character(s). Also called "Containers"
  - SingleQuotes
  - DoubleQuotes
  - Backticks
  - Spoilers
  - ArrowQuote

Order does not generally matter, but if there were two overlapping containers such as in the string "|| spoiler ` monospace || `", order would matter.
It is up to the user to order them appropriately.
"""

# STL
import re
from abc import ABC, abstractmethod
from typing import List, Type

# PDM
import emoji
import regex
from typing_extensions import override

regex.DEFAULT_VERSION = regex.VERSION1


class Preprocessor(ABC):
    @classmethod  # order matters
    @abstractmethod
    def process(cls, msg: str) -> str:
        raise NotImplementedError


class RegexPreprocessor(Preprocessor):
    pattern: "re.Pattern[str]"
    replace: str = " "

    @classmethod
    @override
    def process(cls, msg: str) -> str:
        return re.sub(cls.pattern, cls.replace, msg)


class Regex1Preprocessor(Preprocessor):
    pattern: "regex.Pattern[str]"
    replace: str = " "

    @classmethod
    @override
    def process(cls, msg: str) -> str:
        return regex.sub(cls.pattern, cls.replace, msg)


"""
The following classes are Ignorables.

Ignorables are tokens which do not count toward the accepted number of tokens
or the total number of tokens.
This is generally because they are considered external to Toki Pona.

It is likely that every user will want to use these.
Not having them will cause many false negatives, such as when a URL is divided
into its parts and checked as a token.
"""


class URLs(RegexPreprocessor):
    """Remove http(s) protocol URLs."""

    pattern = re.compile(r"https?:\/\/\S+")


class MarkdownURLs(RegexPreprocessor):
    """Remove URLs in markdown format, replacing them with their corresponding
    text."""

    pattern = re.compile(r"\[(.+?)\]\(https?:\/\/\S+\)")
    replace = r"\1"


class Emails(RegexPreprocessor):
    """Attempt to remove emails, for a particularly strong definition of
    "email".

    https://www.regular-expressions.info/email.html
    """

    pattern = re.compile(
        r"\b[a-zA-Z0-9._%+-]{2,}@[a-zA-Z0-9.-]{2,}\.[a-zA-Z]{2,24}\b",
        flags=re.IGNORECASE,
    )


class Reference(RegexPreprocessor):
    """Remove text contained in double brackets.

    Often used to fetch articles on Wikipedia, or Magic the Gathering
    cards.
    """

    pattern = re.compile(r"\[\[.+\]\]")


class DiscordEmotes(RegexPreprocessor):
    """Remove text-formatted Discord emotes `<flags:name:id>`"""

    pattern = re.compile(r"<a?:[a-zA-Z0-9_]{2,}:[0-9]{2,}>")


class ColonEmotes(RegexPreprocessor):
    """Remove colon-marked emotes `:name:`"""

    pattern = re.compile(r":[a-zA-Z0-9_]{2,}:")


class DiscordMentions(RegexPreprocessor):
    pattern = re.compile(r"<@[\!\&]?[0-9]{2,}>")


class DiscordChannels(RegexPreprocessor):
    pattern = re.compile(r"<#[0-9]{2,}>")


class DiscordSpecial(RegexPreprocessor):
    pattern = re.compile(r"<id:[a-zA-Z0-9_]{4,}>")


class AngleBracketObject(RegexPreprocessor):
    """A generalized version of the Discord-specific angle bracket objects.

    Removes any contiguous (not broken by whitespace) text in angle
    brackets.
    """

    pattern = re.compile(r"<[^<>\s]+>")


"""
The following classes are Containers.

Containers are a special case of Ignorables, where an entire segment of an input
may be removed and not counted toward the accepted or total number of tokens.

Some users may prefer to use these so that they may quote third parties who
would likely be using a language other than Toki Pona.
"""


class SingleQuotes(RegexPreprocessor):
    pattern = re.compile(r"'[^']+'", flags=re.DOTALL)


class DoubleQuotes(RegexPreprocessor):
    pattern = re.compile(r'"[^"]+"', flags=re.DOTALL)


class Backticks(RegexPreprocessor):
    """Remove paired backticks and their contents `like this`"""

    pattern = re.compile(r"`[^`]+`", flags=re.DOTALL)


class Codeblock(RegexPreprocessor):
    """Remove codeblocks marked by a set of three backticks on their own lines.

    Subset of what would be removed by Backticks, but may be preferable.
    """

    pattern = re.compile(
        r"```.+?```",
        flags=re.DOTALL,
    )


class Spoilers(RegexPreprocessor):
    """Remove paired double bars and their contents `||like this||`"""

    pattern = re.compile(r"\|\|(?:(?!\|\|).)+\|\|", flags=re.DOTALL)


class ArrowQuote(RegexPreprocessor):
    """Remove lines beginning with `> `"""

    pattern = re.compile(r"^>\ .+$", re.MULTILINE)


class AllQuotes(RegexPreprocessor):
    pattern = re.compile(
        "|".join(
            [
                SingleQuotes.pattern.pattern,
                DoubleQuotes.pattern.pattern,
                Backticks.pattern.pattern,
                ArrowQuote.pattern.pattern,
            ]
        ),
        flags=re.MULTILINE | re.DOTALL,
    )


class Emoji(Preprocessor):
    @classmethod
    @override
    def process(cls, msg: str) -> str:
        return emoji.replace_emoji(msg)


class ZeroWidths(RegexPreprocessor):
    """Remove the Zero Width Joiner and Zero Width Non-Joiner from the input.

    ZWJ and ZWNJ do serve semantic purposes,
    such as combining many person emojis into the family emojis,
    or ensuring two characters do not become a ligature.
    However, all emojis are considered punctuation by this library,
    so preprocessing ZWJ out is more accurate:
    It will leave behind the component emojis, which will be ignored.

    But ZWJ cannot be considered punctuation for tokenizing purposes because it is used in the middle of words to render them differently.
    In this vein, ZWJ is a function character.

    In the future, it may be smarter to omit ZWJ in the tokenization process,
    or to make the tokenizer smarter by having it keep together collected emojis.
    But in order to do this, emoji would have to be accurately distinguished from all other punctuation.
    """

    pattern = re.compile("[\\U0000200C-\\U0000200D]")


RECOMMENDED_PREPROCESSORS: List[Type[Preprocessor]] = [
    # These are sorted by the "strength" of their definition, which would be roughly
    # "How confidently have we matched this object?"
    # Additionally, MarkdownURLs must come before URLs, and Emoji must come last due to
    # its ability to appear in all of the others.
    Codeblock,
    AngleBracketObject,
    Reference,
    MarkdownURLs,
    URLs,
    Emails,
    Emoji,
]


__all__ = [
    "AllQuotes",
    "AngleBracketObject",
    "ArrowQuote",
    "Backticks",
    "DiscordChannels",
    "DiscordEmotes",
    "DiscordMentions",
    "DiscordSpecial",
    "DoubleQuotes",
    "Emails",
    "Emoji",
    "MarkdownURLs",
    "RECOMMENDED_PREPROCESSORS",
    "Reference",
    "SingleQuotes",
    "Spoilers",
    "URLs",
    "ZeroWidths",
]



# STL
import re
from abc import ABC, abstractmethod
from typing import Set, List

# PDM
import regex
from typing_extensions import override, deprecated

# LOCAL
from sonatoki.utils import regex_escape
from sonatoki.Filters import NimiUCSUR  # seriously this sucks
from sonatoki.constants import (
    ALL_PUNCT,
    INTRA_WORD_PUNCT,
    ALL_SENTENCE_PUNCT,
    UNICODE_WHITESPACE,
    ALL_PUNCT_RANGES_STR,
    UCSUR_CARTOUCHE_LEFT,
    UCSUR_CARTOUCHE_RIGHT,
    UCSUR_MINUS_CARTOUCHE,
)

regex.DEFAULT_VERSION = regex.VERSION1


class Tokenizer(ABC):
    @classmethod
    @abstractmethod
    def tokenize(cls, s: str) -> List[str]: ...


class SetTokenizer(Tokenizer):
    delimiters: Set[str]


class RegexTokenizer(Tokenizer):
    pattern: "re.Pattern[str]"

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [clean for word in re.split(cls.pattern, s) if (clean := word.strip())]


class Regex1Tokenizer(Tokenizer):
    pattern: "regex.Pattern[str]"

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [
            clean for word in regex.split(cls.pattern, s) if (clean := word.strip())
        ]


class WordTokenizer(SetTokenizer):
    delimiters = set(ALL_PUNCT)
    intra_word_punct = set(INTRA_WORD_PUNCT)

    @classmethod
    def is_delimiter(cls, c: str) -> bool:
        return c in cls.delimiters or not c

    @classmethod
    def add_token(cls, s: str, tokens: List[str], last_match: int, i: int):
        if i > last_match:
            tokens.append(s[last_match:i])

    @classmethod
    def to_tokens(cls, s: str) -> List[str]:
        tokens: List[str] = []

        slen = len(s)
        i = 0
        did_skip = False  # ensure exists
        while i < slen:

            # contiguous punctuation chars
            last_match = i
            while i < slen and cls.is_delimiter(s[i]):
                # no special case
                i += 1
            cls.add_token(s, tokens, last_match, i)

            # contiguous writing chars (much harder)
            last_match = i
            while i < slen and not cls.is_delimiter(s[i]):
                did_skip = False
                # we skip and see another writing char, or init

                if NimiUCSUR.filter(s[i]):
                    cls.add_token(s, tokens, last_match, i)
                    tokens.append(s[i])
                    i += 1
                    last_match = i
                    continue

                next_char = s[i + 1] if i + 1 < slen else ""
                if next_char in cls.intra_word_punct:
                    did_skip = True
                    i += 2
                    continue

                i += 1

            if did_skip:
                # we skipped, but there wasn't another writing character
                cls.add_token(s, tokens, last_match, i - 1)
                last_match = i - 1
                # there may be punctuation though
                # TODO: this is duplicated
                while i < slen and cls.is_delimiter(s[i]):
                    i += 1

            cls.add_token(s, tokens, last_match, i)

        return tokens

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        if not s:
            return []

        tokens: List[str] = []
        candidates: List[str] = s.split()

        for candidate in candidates:
            results = cls.to_tokens(candidate)
            tokens.extend(results)

        return tokens


@deprecated(
    "WordTokenizerRe is a previous reference implementation. Its behavior has diverged from WordTokenizer and it may not be restored."
)
class WordTokenizerRe(RegexTokenizer):
    pattern = re.compile(rf"""([{ALL_PUNCT_RANGES_STR}]+|\s+)""")


@deprecated(
    "WordTokenizerRe1 is a previous reference implementation. Its behavior has diverged from WordTokenizer and it may not be restored."
)
class WordTokenizerRe1(Regex1Tokenizer):
    """Reference implementation for WordTokenizer."""

    pattern = regex.compile(r"""([\p{posix_punct}\p{Punctuation}]+|\s+)""")


class SentTokenizer(SetTokenizer):
    delimiters: Set[str] = set(ALL_SENTENCE_PUNCT + "\n")  # regex does \n with a flag
    intra_word_punct: Set[str] = set(INTRA_WORD_PUNCT)
    all_punct: Set[str] = set(ALL_PUNCT + UNICODE_WHITESPACE)

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        if not s:
            return []

        tokens: List[str] = []

        slen = len(s)
        last_match = 0
        i = 0
        while i < slen:
            # if a cartouche appears, we do not want to split on its punctuation
            if s[i] == UCSUR_CARTOUCHE_LEFT:
                right_i = s.find(UCSUR_CARTOUCHE_RIGHT, i)
                contained: set[str] = set()
                if right_i > 0:
                    contained = set(s[i + 1 : right_i])
                # but it must contain only non-cartouche UCSUR chars
                if contained and contained.issubset(UCSUR_MINUS_CARTOUCHE):
                    i = right_i + 1
                    continue
            if s[i] not in cls.delimiters:
                i += 1
                continue
            if s[i] in cls.intra_word_punct:
                prev = s[i - 1] if i > 0 else ""
                next = s[i + 1] if i + 1 < slen else ""
                if (
                    prev
                    and next
                    and prev not in cls.all_punct
                    and next not in cls.all_punct
                ):
                    i += 2
                    continue

            match = s[last_match : i + 1].strip()
            last_match = i + 1  # newlines can strip but idc
            if not match:
                i += 1
                continue
            tokens.append(match)
            i += 1

        match = s[last_match:].strip()
        if match:
            tokens.append(match)

        return tokens


@deprecated(
    "SentTokenizerRe is a previous reference implementation. Its behavior has diverged from SentTokenizer and it may not be restored."
)
class SentTokenizerRe(RegexTokenizer):
    pattern = re.compile(
        rf"""(?<=[{regex_escape(ALL_SENTENCE_PUNCT)}])|$""", flags=re.MULTILINE
    )
    # TODO: are <> or {} that common as *sentence* delims? [] are already a stretch
    # TODO: do the typography characters matter?
    # NOTE: | / and , are *not* sentence delimiters for my purpose


@deprecated(
    "SentTokenizerRe1 is a previous reference implementation. Its behavior has diverged from SentTokenizer and it may not be restored."
)
class SentTokenizerRe1(Regex1Tokenizer):
    pattern = regex.compile(
        rf"""(?<=[{regex_escape(ALL_SENTENCE_PUNCT)}]|$)""", flags=regex.MULTILINE
    )


__all__ = [
    "WordTokenizer",
    "SentTokenizer",
]



# STL
from typing import Dict, List, Union, Literal, TypedDict

Number = Union[int, float]


# TODO: scorecard kinda sucks as a name
class Scorecard(TypedDict):
    text: str
    tokenized: List[str]
    filtered: List[str]
    cleaned: List[str]
    score: Number


LinkuUsageDate = Union[
    Literal["2020-04"],
    Literal["2021-10"],
    Literal["2022-08"],
    Literal["2023-09"],
    Literal["2024-09"],
]

LinkuUsageCategory = Union[
    Literal["core"],
    Literal["common"],
    Literal["uncommon"],
    Literal["obscure"],
    Literal["sandbox"],
]

LinkuBooks = Union[
    Literal["pu"],
    Literal["ku suli"],
    Literal["ku lili"],
    Literal["none"],
]


class LinkuWord(TypedDict):
    id: str
    author_verbatim: str
    author_verbatim_source: str
    book: str
    coined_era: str
    coined_year: str
    creator: List[str]
    ku_data: Dict[str, int]
    see_also: List[str]
    resources: Dict[str, str]
    representations: Dict[str, Union[str, List[str]]]
    source_language: str
    usage_category: LinkuUsageCategory
    word: str
    deprecated: bool
    etymology: List[Dict[str, str]]
    audio: List[Dict[str, str]]
    pu_verbatim: Dict[str, str]
    usage: Dict[LinkuUsageDate, int]
    translations: Dict[str, Dict[str, str]]
